#define _USE_MATH_DEFINES

#include <cmath>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
//using namespace glm;

#include "Camera.h"
#include <iostream>

static const float maximalVerticalAngle = 89.0f;

static inline float convertRadiansToDegrees(float radians) 
{
	return (radians * 180.0f)/(float)M_PI;
}

Camera::Camera() : 
	mPosition(0.0f, 0.0f, 0.0f),
	mHorizontalAngle(0.0f),
	mVerticalAngle(0.0f),
	mFieldOfView(45.0f),
	mNearPlane(0.1f),
	mFarPlane(300.0f),
	mAspectRatio(4.0f/3.0f),
	mOrientation()
{
}

const glm::vec3& Camera::position() const
{
	return mPosition;
}

glm::mat4 Camera::orientation() const
{
	/*glm::mat4 orientation;
	orientation = glm::rotate(mOrientation, mVerticalAngle, glm::vec3(1, 0, 0));
	orientation = glm::rotate(mOrientation, mHorizontalAngle, glm::vec3(0, 1, 0));
	mOrientation = orientation; */
	return mOrientation;
}

void Camera::printAngles()
{
	std::cout<<mVerticalAngle<<" "<<mHorizontalAngle<<std::endl;
}
glm::vec3 Camera::forward() const
{
	glm::vec4 forward = glm::inverse(orientation()) * glm::vec4(0, 0, -1, 1);
	return glm::vec3(forward);
}

glm::vec3 Camera::right() const
{
	glm::vec4 right = glm::inverse(orientation()) * glm::vec4(1, 0, 0, 1);
	return glm::vec3(right);
}

glm::vec3 Camera::up() const
{
	glm::vec4 up = glm::inverse(orientation()) * glm::vec4(0, 1, 0, 1);
	return glm::vec3(up);
}

glm::mat4 Camera::matrix() const
{
	return projection() * view();
}
	
glm::mat4 Camera::projection() const
{
	return glm::perspective(mFieldOfView, mAspectRatio, mNearPlane, mFarPlane);
}
	
glm::mat4 Camera::view() const 
{
	return orientation() * glm::translate(glm::mat4(), -mPosition);
	/* glm::vec3 richtung=glm::vec3(
        sin(mHorizontalAngle)*cos(mVerticalAngle)   //mal cos(neigungswinkel) weil xrichtung im verhaeltnis kuerzer wird
        ,sin(mVerticalAngle)
        ,cos(mHorizontalAngle)*cos(mVerticalAngle)     //mal cos(neigungswinkel) siehe x
      //,0.0f //richtung
    );
	return glm::lookAt(
								mPosition,           // Camera is here
								mPosition+richtung, // and looks there : at the same position, plus "direction"
								glm::vec3(0.0f,1.0f,0.0f)   // Head is up (set to 0,-1,0 to look upside-down)
						   );/**/
}

float Camera::fieldOfView() const
{
	return mFieldOfView;
}
	
float Camera::nearPlane() const
{
	return mNearPlane;
}
	
float Camera::farPlane() const
{
	return mFarPlane;
}
	
float Camera::aspectRatio() const
{
	return mAspectRatio;
}

void Camera::setPosition(const glm::vec3& position)
{
	mPosition = position;
}
	
void Camera::offsetPosition(const glm::vec3& offset)
{
	mPosition += offset;
}
void Camera::setFieldOfView(float fieldOfView)
{
	if (fieldOfView > 0.0f && fieldOfView < 180.0f)
		mFieldOfView = fieldOfView;
}
	
void Camera::setPlanes(float nearPlane, float farPlane)
{
	if (nearPlane > 0.0f && farPlane > nearPlane)
	{
		mNearPlane = nearPlane;
		mFarPlane = farPlane;
	}
}

void Camera::setOrientation(glm::mat4 o) 
{
	mOrientation = o;
}
void Camera::setOrientation(float up, float right)
{
	mHorizontalAngle = right;
	mVerticalAngle = up;
	normalizeAngles();
	glm::mat4 orientation;
    orientation = glm::rotate(orientation, mVerticalAngle, glm::vec3(1,0,0));
    orientation = glm::rotate(orientation, mHorizontalAngle, glm::vec3(0,1,0));
    setOrientation(orientation);
	
}
void Camera::offsetOrientation(float up, float right)
{
	setOrientation(up+mVerticalAngle,right+mHorizontalAngle);
}

void Camera::lookAt(glm::vec3 position)
{
	if (position != mPosition)
	{
		glm::vec3 direction = glm::normalize(position - mPosition);
		mVerticalAngle = convertRadiansToDegrees(asinf(-direction.y));
		mHorizontalAngle = -convertRadiansToDegrees(atan2f(-direction.x, -direction.z));
		normalizeAngles();
	}
}

void Camera::setAspectRatio(float aspectRatio)
{
	if (aspectRatio > 0.0f)
		mAspectRatio = aspectRatio;
}

void Camera::normalizeAngles() 
{
	mHorizontalAngle = fmodf(mHorizontalAngle, 360.0f);

	while (mHorizontalAngle < 0.0f) 
		mHorizontalAngle += 360.0f;

	if (mVerticalAngle > maximalVerticalAngle)
		mVerticalAngle = maximalVerticalAngle;
	else if (mVerticalAngle < -maximalVerticalAngle)
		mVerticalAngle = -maximalVerticalAngle;
}

Camera::~Camera()
{
}
